//                                               -*- C++ -*-
/**
 *  @file  ComposedNumericalMathHessianImplementation.cxx
 *  @brief The class that implements the composition between numerical math hessians
 *
 *  (C) Copyright 2005-2007 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2008-10-31 11:52:04 +0100 (ven 31 oct 2008) $
 *  Id:      $Id: ComposedNumericalMathHessianImplementation.cxx 995 2008-10-31 10:52:04Z dutka $
 */
#include "ComposedNumericalMathHessianImplementation.hxx"
#include "PersistentObjectFactory.hxx"

namespace OpenTURNS {

  namespace Base {

    namespace Func {

      CLASSNAMEINIT(ComposedNumericalMathHessianImplementation);

      static Common::Factory<ComposedNumericalMathHessianImplementation> RegisteredFactory("ComposedNumericalMathHessianImplementation");

      /* Default constructor */
      ComposedNumericalMathHessianImplementation::ComposedNumericalMathHessianImplementation(const GradientImplementation & p_leftGradient,
											     const HessianImplementation & p_leftHessian,
											     const EvaluationImplementation & p_rightFunction,
											     const GradientImplementation & p_rightGradient,
											     const HessianImplementation & p_rightHessian)
	throw(InvalidArgumentException)
	: NumericalMathHessianImplementation(),
	  p_leftGradient_(p_leftGradient),
	  p_leftHessian_(p_leftHessian),
	  p_rightFunction_(p_rightFunction),
	  p_rightGradient_(p_rightGradient),
	  p_rightHessian_(p_rightHessian)
      {
	UnsignedLong p_rightInputDimension(p_rightFunction->getInputNumericalPointDimension());
	UnsignedLong p_rightOutputDimension(p_rightFunction->getOutputNumericalPointDimension());
	UnsignedLong p_leftInputDimension(p_leftHessian->getInputNumericalPointDimension());
	UnsignedLong p_leftOutputDimension(p_leftHessian->getOutputNumericalPointDimension());
	// We check that the dimensions of the input parameters are compatible
	if ((p_leftInputDimension != p_rightOutputDimension) || (p_leftGradient->getInputNumericalPointDimension() != p_leftInputDimension) || (p_leftGradient->getOutputNumericalPointDimension() != p_leftOutputDimension) || (p_rightGradient->getInputNumericalPointDimension() != p_rightInputDimension) || (p_rightGradient->getOutputNumericalPointDimension() != p_rightOutputDimension) || (p_rightHessian->getInputNumericalPointDimension() != p_rightInputDimension) || (p_rightHessian->getOutputNumericalPointDimension() != p_rightOutputDimension)) throw InvalidArgumentException(HERE) << "The left gradient dimensions are not compatible with the right function dimensions, or the right gradient dimensions are not compatible with the right function dimensions";
      }

      /* Virtual constructor */
      ComposedNumericalMathHessianImplementation * ComposedNumericalMathHessianImplementation::clone() const
      {
	return new ComposedNumericalMathHessianImplementation(*this);
      }

      /* Comparison operator */
      Bool ComposedNumericalMathHessianImplementation::operator ==(const ComposedNumericalMathHessianImplementation & other) const
      {
	return true;
      }
  
      /* String converter */
      String ComposedNumericalMathHessianImplementation::str() const {
	OSS oss;
	oss << "class=" << ComposedNumericalMathHessianImplementation::GetClassName()
	    << " name=" << getName()
	    << " leftGradient=" << p_leftGradient_->str()
	    << " leftHessian=" << p_leftHessian_->str()
	    << " rightFunction=" << p_rightFunction_->str()
	    << " rightGradient=" << p_rightGradient_->str()
	    << " rightHessian=" << p_rightHessian_->str();
	return oss;
      }
  
      /* Method hessian() returns the symetric tensor of the second gradient of the function at point */
      ComposedNumericalMathHessianImplementation::SymmetricTensor ComposedNumericalMathHessianImplementation::hessian(const NumericalPoint & in) const
	throw(InvalidArgumentException,InternalException)
      {
	callsNumber_++;
	// Hessian of the composed function fog, g:R^n->R^p, f:R^p->R^q so fog:R^n->R^q
	NumericalPoint p_rightIn(p_rightFunction_->operator()(in));
	Matrix p_rightGradientIn(p_rightGradient_->gradient(in));
	SymmetricTensor p_rightHessianIn(p_rightHessian_->hessian(in));
	Matrix p_leftGradientIn(p_leftGradient_->gradient(p_rightIn));
	SymmetricTensor p_leftHessianIn(p_leftHessian_->hessian(p_rightIn));
	// Get the several dimensions
	UnsignedLong inputDimension(getInputNumericalPointDimension());
	UnsignedLong outputDimension(getOutputNumericalPointDimension());
	UnsignedLong innerDimension(p_rightIn.getDimension());
	SymmetricTensor result(inputDimension, outputDimension);
	/* We unroll the formula:
	 *  D2(h)(x) = D2(f)(g(x))D(g)(x)D(g)(x)+D(f)(g(x))D2(g)(x)
	 *  after transposition to convert jacobian and hessian into
	 *  the associated gradients
	 */
	for (UnsignedLong i = 0; i < inputDimension; i++)
	  {
	    for (UnsignedLong j = 0; j <= i; j++)
	      {
		for (UnsignedLong k = 0; k < outputDimension; k++)
		  {
		    result(i, j, k) = 0.0;
		    for (UnsignedLong m = 0; m < innerDimension; m++)
		      {
			result(i, j, k) += p_leftGradientIn(m, k) * p_rightHessianIn(i, j, m);
			for (UnsignedLong n = 0; n < p_rightIn.getDimension(); n++)
			  {
			    result(i, j, k) += p_leftHessianIn(m, n, k) * p_rightGradientIn(i, m) * p_rightGradientIn(j, n);
			  } // n
		      } // m
		  } // k
	      } // j
	  } // i
	return result;
      }

      /* Accessor for input point dimension */
      UnsignedLong ComposedNumericalMathHessianImplementation::getInputNumericalPointDimension() const
	throw(InternalException)
      {
	return p_rightHessian_->getInputNumericalPointDimension();
      }
      
      /* Accessor for output point dimension */
      UnsignedLong ComposedNumericalMathHessianImplementation::getOutputNumericalPointDimension() const
	throw(InternalException)
      {
	return p_leftHessian_->getOutputNumericalPointDimension();
      }
     
      /* Method save() stores the object through the StorageManager */
      void ComposedNumericalMathHessianImplementation::save(const StorageManager::Advocate & adv) const
      {
	NumericalMathHessianImplementation::save(adv);
	adv.writeValue(*p_leftGradient_, StorageManager::MemberNameAttribute, "leftGradient_");
	adv.writeValue(*p_leftHessian_, StorageManager::MemberNameAttribute, "leftHessian_");
	adv.writeValue(*p_rightFunction_, StorageManager::MemberNameAttribute, "rightFunction_");
	adv.writeValue(*p_rightGradient_, StorageManager::MemberNameAttribute, "rightGradient_");
	adv.writeValue(*p_rightHessian_, StorageManager::MemberNameAttribute, "rightHessian_");
      }

      /* Method load() reloads the object from the StorageManager */
      void ComposedNumericalMathHessianImplementation::load(const StorageManager::Advocate & adv)
      {
	NumericalMathHessianImplementation::load(adv);
	Common::TypedInterfaceObject<NumericalMathEvaluationImplementation> evaluationValue;
	Common::TypedInterfaceObject<NumericalMathGradientImplementation> gradientValue;
	Common::TypedInterfaceObject<NumericalMathHessianImplementation> hessianValue;
	  
	StorageManager::List objList = adv.getList(StorageManager::ObjectEntity);
	if (objList.readValue(evaluationValue,  StorageManager::MemberNameAttribute, "rightFunction_"))
	  p_rightFunction_ = evaluationValue.getImplementation();
	if (objList.readValue(gradientValue, StorageManager::MemberNameAttribute, "leftGradient_"))
	  p_leftGradient_ = gradientValue.getImplementation();
	if (objList.readValue(gradientValue, StorageManager::MemberNameAttribute, "rightGradient_"))
	  p_rightGradient_ = gradientValue.getImplementation();
	if (objList.readValue(hessianValue, StorageManager::MemberNameAttribute, "leftHessian_"))
          p_leftHessian_ = hessianValue.getImplementation();
	if (objList.readValue(hessianValue, StorageManager::MemberNameAttribute, "rightHessian_"))
	  p_rightHessian_= hessianValue.getImplementation();
      }

    } /* namespace Func */
  } /* namespace Base */
} /* namespace OpenTURNS */
