//                                               -*- C++ -*-
/**
 * @file  ComputedNumericalMathEvaluationImplementation.cxx
 * @brief Abstract top-level class for all distributions
 *
 * (C) Copyright 2005-2006 EDF
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 *
 * \author $LastChangedBy: dutka $
 * \date   $LastChangedDate: 2008-10-31 11:52:04 +0100 (ven 31 oct 2008) $
 */

#include "ComputedNumericalMathEvaluationImplementation.hxx"
#include "ComputedNumericalMathEvaluationImplementationFactory.hxx"
#include "PersistentObjectFactory.hxx"
#include "Cache.hxx"
#include "Log.hxx"

namespace OpenTURNS {

  namespace Base {

    namespace Func {

      using Common::Log;

      CLASSNAMEINIT(ComputedNumericalMathEvaluationImplementation);

      static Common::Factory<ComputedNumericalMathEvaluationImplementation> RegisteredFactory("ComputedNumericalMathEvaluationImplementation");

      /* Default constructor */
      ComputedNumericalMathEvaluationImplementation::ComputedNumericalMathEvaluationImplementation(const String & name,
												   const WrapperFile & file)
	throw(WrapperInternalException)	:
	NumericalMathEvaluationImplementation(),
	p_function_(0),
	p_state_(0)
      {
	setName(name);
	if (file.getWrapperData().getFunctionDescription().provided_) {
	  p_function_.reset(new WrapperObject( file.getWrapperData().getLibraryPath(),
					       file.getWrapperData().getFunctionDescription().name_,
					       file.getWrapperData(),
					       WrapperObject::FUNCTION ));
	}

	if (p_function_.isNull()) throw WrapperInternalException(HERE) << "Unable to allocate wrapper";

	// Initialize the state into the wrapper
	p_state_ = p_function_->createNewState();
	NumericalMathEvaluationImplementation::setDescription(getDescription());

	// Activate the cache because external codes are supposed expensive
	NumericalMathEvaluationImplementation::enableCache();
      }


      /* Copy constructor */
      ComputedNumericalMathEvaluationImplementation::ComputedNumericalMathEvaluationImplementation(const ComputedNumericalMathEvaluationImplementation & other)
	throw(WrapperInternalException)
	: NumericalMathEvaluationImplementation(other),
	  p_function_(other.p_function_),
	  p_state_(0)
      {
	if (p_function_.isNull()) throw WrapperInternalException(HERE) << "Unable to allocate wrapper";

	// Initialize the state into the wrapper
	p_state_ = p_function_->createNewState();
      }


      /* Virtual constructor */
      ComputedNumericalMathEvaluationImplementation * ComputedNumericalMathEvaluationImplementation::clone() const
      {
	return new ComputedNumericalMathEvaluationImplementation(*this);
      }

      /* Destructor */
      ComputedNumericalMathEvaluationImplementation::~ComputedNumericalMathEvaluationImplementation()
      {
	// Call the finalization function before destruction
	p_function_->finalize( p_state_ );

	// Delete the state into the wrapper
	p_function_->deleteState( p_state_ );
      }


      /* Comparison operator */
      Bool ComputedNumericalMathEvaluationImplementation::operator ==(const ComputedNumericalMathEvaluationImplementation & other) const
      {
	return true;
      }

  
      /* String converter */
      String ComputedNumericalMathEvaluationImplementation::str() const {
	OSS oss;
	oss << "class=" << ComputedNumericalMathEvaluationImplementation::GetClassName()
	    << " name=" << getName();
	return oss;
      }

  



      /* State accessor */
      void * ComputedNumericalMathEvaluationImplementation::getState() const
      {
	return p_state_;
      }






      /* Here is the interface that all derived class must implement */
	
      /* Operator () */
      ComputedNumericalMathEvaluationImplementation::NumericalPoint
      ComputedNumericalMathEvaluationImplementation::operator() (const NumericalPoint & in) const
	throw(InvalidArgumentException,InternalException)
      {
	try {
	  // First, initialize the external code on first invocation
	  if (callsNumber_ == 0) p_function_->initialize( p_state_ );

	  // Execute the wrapper
	  if ( p_cache_->hasKey(in) ) return p_cache_->find(in);
	  else {
	    ++callsNumber_;
	    NumericalPoint out = p_function_->execute( p_state_, in );
	    if (p_cache_->isEnabled()) p_cache_->add( in, out );
	    return out;
	  }
	}
	catch (InvalidArgumentException & ex) {
	  throw;
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex.str();
	}
      }

      /* Operator () */
      ComputedNumericalMathEvaluationImplementation::NumericalSample
      ComputedNumericalMathEvaluationImplementation::operator() (const NumericalSample & in) const
	throw(InvalidArgumentException, InternalException)
      {

	try {
	  // First, initialize the external code on first invocation
	  if (callsNumber_ == 0) p_function_->initialize( p_state_ );

	  if (p_cache_->isEnabled()) {
	    // We build a new sample without the already computed points
	    NumericalSample newIn( 0, in.getDimension() );
	    for(UnsignedLong i=0; i<in.getSize(); ++i) {
	      if ( ! p_cache_->hasKey( in[i] ) ) newIn.add( in[i] );
	    }


	    // Execute the wrapper if needed
	    NumericalSample newOut( 0, getOutputNumericalPointDimension() );
	    if (newIn.getSize() > 0) {

	      // We shrink newIn and remove double entries
	      std::sort( newIn.begin(), newIn.end() );
	      NumericalSample::iterator last = std::unique( newIn.begin(), newIn.end() );
	      newIn.erase( last, newIn.end() );
	    
	      callsNumber_ += newIn.getSize();
	      newOut = p_function_->execute( p_state_, newIn );
	    }

	    // We use a secondary cache to hold the computed values
	    Type::Cache tempCache( newIn.getSize() );
	    tempCache.enable();
	    for(UnsignedLong i=0; i<newIn.getSize(); ++i) tempCache.add( newIn[i], newOut[i] );

	    // We gather the computed values into a new output sample
	    NumericalSample out( in.getSize(), getOutputNumericalPointDimension() );
	    for(UnsignedLong i=0; i<in.getSize(); ++i) {
	      if ( p_cache_->hasKey(in[i]) ) {
		out[i] = p_cache_->find( in[i] );

	      } else if ( tempCache.hasKey(in[i]) ) {
		out[i] = tempCache.find( in[i] );

	      } else throw InternalException(HERE) << "Error in computing sample. Some values can not be retrieved from cache.";
	    
	    }

	    // We add the computed values into the cache AFTER having read the cache because
	    // older values may be flushed
	    //for(UnsignedLong i=0; i<newIn.getSize(); ++i) p_cache_->add( newIn[i], newOut[i] );
	    p_cache_->merge( tempCache );

	    // return the gathered sample
	    return out;

	  } else {
	    callsNumber_ += in.getSize();
	    return p_function_->execute( p_state_, in );
	  }

	}
	catch (InvalidArgumentException & ex) {
	  throw;
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex.str();
	}
      }



      /* Accessor for input point dimension */
      UnsignedLong ComputedNumericalMathEvaluationImplementation::getInputNumericalPointDimension() const
	throw(InternalException)
      {
	UnsignedLong inputDimension = 0;

	try {
	  inputDimension = p_function_->getInNumericalPointDimension( p_state_ );
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex.str();
	}

	return inputDimension;
      }



      /* Accessor for output point dimension */
      UnsignedLong ComputedNumericalMathEvaluationImplementation::getOutputNumericalPointDimension() const
	throw(InternalException)
      {
	UnsignedLong outputDimension = 0;

	try {
	  outputDimension = p_function_->getOutNumericalPointDimension( p_state_ );
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex.str();
	}

	return outputDimension;
      }

      /* Accessor for output point dimension */
      ComputedNumericalMathEvaluationImplementation::Description ComputedNumericalMathEvaluationImplementation::getDescription() const
	throw(InternalException)
      {
	Description description;

	try {
	  description = p_function_->getDescription();
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex.str();
	}

	return description;
      }
      
      /* Method save() stores the object through the StorageManager */
      void ComputedNumericalMathEvaluationImplementation::save(const StorageManager::Advocate & adv) const
      {
	NumericalMathEvaluationImplementation::save(adv);
      }

      /* Method load() reloads the object from the StorageManager */
      void ComputedNumericalMathEvaluationImplementation::load(const StorageManager::Advocate & adv)
      {
	NumericalMathEvaluationImplementation::load(adv);
	*this = *static_cast<ComputedNumericalMathEvaluationImplementation*>(ComputedNumericalMathEvaluationImplementationFactory::getInstance()->buildObject(getName()));
      }

    } /* namespace Func */
  } /* namespace Base */
} /* namespace OpenTURNS */
