//                                               -*- C++ -*-
/**
 *  @file  LinearModelFactory.cxx
 *  @brief Factory for linear model
 *
 *  (C) Copyright 2005-2007 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2007-09-10 15:54:48 +0200 (lun 10 sep 2007) $
 *  Id:      $Id: LinearModelFactory.cxx 503 2007-09-10 13:54:48Z dutka $
 */
#include <fstream>
#include "LinearModelFactory.hxx"
#include "LinearModel.hxx"
#include "Path.hxx"
#include "ResourceMap.hxx"
#include "Exception.hxx"
//#include "RBase.hxx"


namespace OpenTURNS
{

  namespace Base
  {

    namespace Stat
    {

      const NumericalScalar LinearModelFactory::DefaultLevelValue;

      //typedef Interface::RBase RBase;
      typedef LinearModel::ConfidenceIntervalPersistentCollection ConfidenceIntervalPersistentCollection;
      typedef LinearModel::NumericalScalarPersistentCollection    NumericalScalarPersistentCollection;
      typedef LinearModel::NumericalPoint                         NumericalPoint;
      typedef Common::Path                                        Path;
      typedef Common::Exception                                   InternalException;

      /* Constructor */
      LinearModelFactory::LinearModelFactory()
      {
	// Nothing to do
      }

      /* Destructor */
      LinearModelFactory::~LinearModelFactory()
      {
	// Nothing to do
      }

	
      /* LM creation */
      LinearModel LinearModelFactory::buildLM(const NumericalSample & samplePred,
                                              const NumericalSample & sampleLab,
                                              const NumericalScalar levelValue) const throw(InvalidArgumentException)
      {
       	if (samplePred.getSize() != sampleLab.getSize()) throw InvalidArgumentException(HERE) << "Error: predictors sample have the same size than the laboratory sample";
	String predictorFileName(samplePred.storeToTemporaryFile());
	String laboratoryFileName(sampleLab.storeToTemporaryFile());
	String resultFileName(Path::BuildTemporaryFileName("RResult.txt.XXXXXX"));
	String commandFileName(Path::BuildTemporaryFileName("RCmd.R.XXXXXX"));
        std::ofstream cmdFile(commandFileName.c_str(), std::ios::out);
	// Fill-in the command file
        cmdFile << "library(rotRPackage)" << std::endl;
        cmdFile << "options(digits=17)" << std::endl;
        cmdFile << "options(warn=-1)" << std::endl;
        cmdFile << "samplePred <- data.matrix(read.table(\"" << predictorFileName << "\"))" << std::endl;
	cmdFile << "sampleLab <- data.matrix(read.table(\"" << laboratoryFileName << "\"))" << std::endl;
	cmdFile << "res <- computeLinearModel(samplePred, sampleLab, " << levelValue <<")" << std::endl;
        cmdFile << "f <- file(\"" << resultFileName << "\",\"wt\")" << std::endl;
        cmdFile << "cat(res$parameterEstimate, res$confidenceIntervalLow, res$confidenceIntervalHigh, res$pValues, sep=\"\\n\", file=f)" << std::endl;
        cmdFile << "close(f)" << std::endl;
        cmdFile.close();

	OSS systemCommand;
	systemCommand << Common::ResourceMap::GetInstance().get("R-executable-command") << " --no-save --silent < " << commandFileName << " 2>&1 > /dev/null";
	int returnCode(system(String(systemCommand).c_str()));
        if (returnCode != 0) throw InternalException(HERE) << "Error: unable to execute the system command " << String(systemCommand) << " returned code is " << returnCode;
	// Parse result file
        std::ifstream resultFile(resultFileName.c_str(), std::ios::in);
	UnsignedLong dimension(samplePred.getDimension() + 1);
	NumericalPoint regression(dimension);
	// Read the regression parameters
	for (UnsignedLong i = 0; i < dimension; i++)
	  {
	    resultFile >> regression[i];
	  }
	// Read the lower bounds of the intervals
	NumericalPoint lowerBounds(dimension);
	for (UnsignedLong i = 0; i < dimension; i++)
	  {
	    resultFile >> lowerBounds[i];
	  }
	// Read the upper bounds of the intervals
	NumericalPoint upperBounds(dimension);
	for (UnsignedLong i = 0; i < dimension; i++)
	  {
	    resultFile >> upperBounds[i];
	  }
	// Convert the bounds to an interval collection
	ConfidenceIntervalPersistentCollection confidenceIntervals(dimension, ConfidenceInterval(0.0, 1.0));
	for (UnsignedLong i = 0; i < dimension; i++)
	  {
	    confidenceIntervals[i].setValues(lowerBounds[i], upperBounds[i]);
	  }
	// Read the p-values of the coefficients
	NumericalScalarPersistentCollection pValues(dimension);
	for (UnsignedLong i = 0; i < dimension; i++)
	  {
	    resultFile >> pValues[i];
	  }
	// We only remove those files for low debug levels
#if OT_DEBUG_LEVEL < 6
	if (remove(predictorFileName.c_str()) == -1) throw InternalException(HERE) << "Error: cannot remove file " << predictorFileName;
	if (remove(laboratoryFileName.c_str()) == -1) throw InternalException(HERE) << "Error: cannot remove file " << laboratoryFileName;;
	if (remove(resultFileName.c_str()) == -1) throw InternalException(HERE) << "Error: cannot remove file " << resultFileName;
	if (remove(commandFileName.c_str()) == -1) throw InternalException(HERE) << "Error: cannot remove file " << commandFileName;
#endif
	return LinearModel(regression, confidenceIntervals, pValues);
	//	return RBase(CONNECTOR_ROT).computeLinearModel(samplePred, sampleLab, levelValue);
      }

    } /* namespace Stat */
  } /* namespace Base */
} /* namespace OpenTURNS */

