//                                               -*- C++ -*-
/**
 *  @file  Exponential.cxx
 *  @brief The Exponential distribution
 *
 *  (C) Copyright 2005-2007 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2008-09-13 22:37:56 +0200 (sam 13 sep 2008) $
 *  Id:      $Id: Exponential.cxx 929 2008-09-13 20:37:56Z dutka $
 */
#include <cmath>
#include "Exponential.hxx"
#include "SpecFunc.hxx"
#include "RandomGenerator.hxx"
#include "PersistentObjectFactory.hxx"

namespace OpenTURNS {

  namespace Uncertainty {

    namespace Distribution {

      typedef Base::Stat::RandomGenerator RandomGenerator;

      CLASSNAMEINIT(Exponential);

      static Base::Common::Factory<Exponential> RegisteredFactory("Exponential");

      /* Default constructor */
      Exponential::Exponential()
	: NonEllipticalDistribution("Exponential"),
	  lambda_(1.0),
	  gamma_(0.0)
      {
	setDimension( 1 );
	computeRange();
      }

      /* Parameters constructor */
      Exponential::Exponential(const NumericalScalar lambda,
			       const NumericalScalar gamma)
	throw(InvalidArgumentException)
	: NonEllipticalDistribution("Exponential"),
	  lambda_(lambda),
	  gamma_(gamma)
      {
	// We set the dimension of the Exponential distribution
	setDimension( 1 );
	computeRange();
      }

      /* Comparison operator */
      Bool Exponential::operator ==(const Exponential & other) const
      {
	Bool sameObject = false;

	if (this != &other) { // Other is NOT me, so I have to realize the comparison
	  if ( (lambda_ == other.lambda_) && (gamma_ == other.gamma_) )
	    sameObject = true;

	} else sameObject = true;

	return sameObject;
      }
  
      /* String converter */
      String Exponential::str() const
      {
	OSS oss;
	oss << "class=" << Exponential::GetClassName()
	    << " name=" << getName()
	    << " dimension=" << getDimension()
	    << " lambda=" << lambda_
	    << " gamma=" << gamma_;
	return oss;
      }
  
      /* Virtual constructor */
      Exponential * Exponential::clone() const
      {
	return new Exponential(*this);
      }

      /* Compute the numerical range of the distribution given the parameters values */
      void Exponential::computeRange()
      {
	const NumericalPoint lowerBound(1, gamma_);
	const NumericalPoint upperBound(1, gamma_ + (log(lambda_) + SpecFunc::LogMinNumericalScalar) / lambda_);
	const Interval::BoolCollection finiteLowerBound(1, true);
	const Interval::BoolCollection finiteUpperBound(1, false);
	setRange(Interval(lowerBound, upperBound, finiteLowerBound, finiteUpperBound));
      }

      /* Get one realization of the distribution */
      Exponential::NumericalPoint Exponential::getRealization() const
      {
	return NumericalPoint(1, gamma_ - log(RandomGenerator::Generate()) / lambda_);
      }
     

      /* Get the DDF of the distribution */
      Exponential::NumericalPoint Exponential::computeDDF(const NumericalPoint & point) const
      {
	if (point[0] < gamma_) return NumericalPoint(1, 0.0);
	return NumericalPoint(1, -lambda_ * computePDF(point));
      }


      /* Get the PDF of the distribution */
      NumericalScalar Exponential::computePDF(const NumericalPoint & point) const
      {
	NumericalScalar x(point[0] - gamma_);
	if (x < 0.0) return 0.0;
	return lambda_ * exp(-lambda_ * x);
      }


      /* Get the CDF of the distribution */
      NumericalScalar Exponential::computeCDF(const NumericalPoint & point, const Bool tail) const
      {
	NumericalScalar x(point[0] - gamma_);
	if (x < 0.0) return (tail ? 1.0 : 0.0);
	if (tail) return exp(-lambda_ * x);
	return 1.0 - exp(-lambda_ * x);
      }

      /* Get the characteristic function of the distribution, i.e. phi(u) = E(exp(I*u*X)) */
      NumericalComplex Exponential::computeCharacteristicFunction(const NumericalPoint & point) const
      {
	NumericalScalar u(point[0]);
	return lambda_ * exp(NumericalComplex(0, -u * gamma_)) / NumericalComplex(lambda_, -u);
      }

      /* Get the PDFGradient of the distribution */
      Exponential::NumericalPoint Exponential::computePDFGradient(const NumericalPoint & point) const
      {
	NumericalScalar x(point[0] - gamma_);
	NumericalPoint pdfGradient(2, 0.0);
	if (x < 0.0) return pdfGradient;
	NumericalScalar expX(exp(-lambda_ * x));
	pdfGradient[0] = (1.0 - lambda_ * x) * expX;
	pdfGradient[1] = lambda_ * lambda_ * expX;
	return pdfGradient;
      }

      /* Get the CDFGradient of the distribution */
      Exponential::NumericalPoint Exponential::computeCDFGradient(const NumericalPoint & point) const
      {
	NumericalScalar x(point[0] - gamma_);
	NumericalPoint cdfGradient(2, 0.0);
	if (x < 0.0) return cdfGradient;
	NumericalScalar expX(exp(-lambda_ * x));
	cdfGradient[0] = x * expX;
	cdfGradient[1] = -lambda_ * expX;
	return cdfGradient;
      }

      /* Get the quantile of the distribution */
      Exponential::NumericalPoint Exponential::computeQuantile(const NumericalScalar prob) const
      {
	if (prob < 0.0 || prob > 1.0) throw InvalidArgumentException(HERE) << "Error: cannot compute a quantile for a probability level outside of [0, 1]";
	// Special case for boarding values
	if (prob == 0.0) return getRange().getLowerBound();
	if (prob == 1.0) return getRange().getUpperBound();
	return NumericalPoint(1, gamma_ - log(1.0 - prob) / lambda_);
      }

      /* Get the mean of the distribution */
      Exponential::NumericalPoint Exponential::getMean() const throw(NotDefinedException)
      {
	return NumericalPoint(1, gamma_ + 1.0 / lambda_);
      }

      /* Get the standard deviation of the distribution */
      Exponential::NumericalPoint Exponential::getStandardDeviation() const throw(NotDefinedException)
      {
	return NumericalPoint(1, 1.0 / lambda_);
      }

      /* Get the skewness of the distribution */
      Exponential::NumericalPoint Exponential::getSkewness() const throw(NotDefinedException)
      {
	return NumericalPoint(1, 2.0);
      }

      /* Get the kurtosis of the distribution */
      Exponential::NumericalPoint Exponential::getKurtosis() const throw(NotDefinedException)
      {
	return NumericalPoint(1, 9.0);
      }

      /* Get the covariance of the distribution */
      Exponential::CovarianceMatrix Exponential::getCovariance() const throw(NotDefinedException)
      {
	CovarianceMatrix covariance(1);
	covariance(0, 0) = 1.0 / (lambda_ * lambda_);
	return covariance;
      }

      /* Parameters value and description accessor */
      Exponential::NumericalPointWithDescriptionCollection Exponential::getParametersCollection() const
      {
	NumericalPointWithDescriptionCollection parameters(1);
	NumericalPointWithDescription point(2);
        Description description(point.getDimension());
	point[0] = lambda_;
	point[1] = gamma_;
	description[0] = "lambda";
	description[1] = "gamma";
	point.setDescription(description);
	point.setName(getDescription()[0]);
	parameters[0] = point;
	return parameters;
      }

      /* Lambda accessor */
      void Exponential::setLambda(const NumericalScalar lambda)
	throw(InvalidArgumentException)
      {
	if (lambda <= 0.) throw InvalidArgumentException(HERE) << "Lambda MUST be positive";
	lambda_ = lambda;
	computeRange();
      }

      /* Lambda accessor */
      NumericalScalar Exponential::getLambda() const
      {
	return lambda_;
      }


      /* Gamma accessor */
      void Exponential::setGamma(const NumericalScalar gamma)
      {
	gamma_ = gamma;
	computeRange();
      }

      /* Gamma accessor */
      NumericalScalar Exponential::getGamma() const
      {
	return gamma_;
      }

      /* Method save() stores the object through the StorageManager */
      void Exponential::save(const StorageManager::Advocate & adv) const
      {
	NonEllipticalDistribution::save(adv);
	adv.writeValue("lambda_", lambda_);
	adv.writeValue("gamma_", gamma_);
      }

      /* Method load() reloads the object from the StorageManager */
      void Exponential::load(const StorageManager::Advocate & adv)
      {
	NonEllipticalDistribution::load(adv);

	String name;
	NumericalScalar value;
	StorageManager::List objList = adv.getList(StorageManager::NumericalScalarEntity);
	for(objList.firstValueToRead(); objList.moreValuesToRead(); objList.nextValueToRead()) {
	  if (objList.readValue(name, value)) {
	    if (name == "lambda_") lambda_ = value;
	    if (name == "gamma_") gamma_ = value;
	  }
	}
	computeRange();
      }
      



    } /* namespace Distribution */
  } /* namespace Uncertainty */
} /* namespace OpenTURNS */
