//                                               -*- C++ -*-
/**
 *  @file  StandardEvent.cxx
 *  @brief The class that implements an event based on a composite random vector whose
 *
 *  (C) Copyright 2005-2007 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2008-09-22 11:34:11 +0200 (lun 22 sep 2008) $
 *  Id:      $Id: StandardEvent.cxx 941 2008-09-22 09:34:11Z dutka $
 */
#include "StandardEvent.hxx"
#include "NumericalMathFunction.hxx"
#include "Distribution.hxx"
#include "CompositeRandomVector.hxx"
#include "UsualRandomVector.hxx"
#include "RandomVectorImplementation.hxx"
#include "Pointer.hxx"
#include "ComparisonOperatorImplementation.hxx"

namespace OpenTURNS {

  namespace Uncertainty {

    namespace Model {

      typedef Base::Func::NumericalMathFunction                   NumericalMathFunction;
      typedef Distribution::InverseIsoProbabilisticTransformation InverseIsoProbabilisticTransformation;

      CLASSNAMEINIT(StandardEvent);

      /* Constructor from a RandomVector */
      StandardEvent::StandardEvent(const RandomVector & antecedent,
				   const ComparisonOperator & op,
				   const NumericalScalar threshold,
				   const String & name)
	throw (InvalidArgumentException)
	: Event(antecedent, op, threshold, name)
      {
	// StandardEvent can only be constructed from composite random vector whose antecedent has a spherical distribution. As we cannot check it, we just check that the distribution is elliptical
	if (!antecedent.getImplementation()->getAntecedent()->getDistribution().getImplementation()->isElliptical()) throw InvalidArgumentException(HERE) << "StandardEvent can only be constructed from composite random vectors whose antecedent is elliptical, here the distribution is " << antecedent.getImplementation()->getAntecedent()->getDistribution().getImplementation()->str();
      }
     
      /* Constructor from an Event */
      StandardEvent::StandardEvent(const Event & event,
                                   const String & name)
	throw (InvalidArgumentException)
	: Event(event)
      {
	// Get the isoprobabilistic inverse transformation from the antecedent distribution
	Distribution distribution(event.getImplementation()->getAntecedent()->getDistribution());
	InverseIsoProbabilisticTransformation inverse(distribution.getInverseIsoProbabilisticTransformation());
	// Get the function
	NumericalMathFunction function(event.getImplementation()->getFunction());
	// Get the standard distribution from the antecedent distribution
	Distribution standard(distribution.getStandardDistribution());
	// Build the antecedent
	RandomVector antecedent(standard);
	NumericalMathFunction composed(function, inverse);
	RandomVector vect(composed, antecedent);
	// Set the random vector implementation
	(*this) = StandardEvent(vect, event.getOperator(), event.getThreshold(), event.getName());
      }

    } /* namespace Model */
  } /* namespace Uncertainty */
} /* namespace OpenTURNS */
